#!/usr/bin/env python3
# coding: utf-8


def numsys_cast(num, base, precision=0):
    if not isinstance(base, int) or base <= 1:
        raise TypeError('base must be an integer greater than 1')

    integer, fractional = divmod(num, 1)
    integer = int(integer)
    integer_digits = []
    fractional_digits = []

    while integer:
        integer, digi = divmod(integer, base)
        integer_digits.append(digi)
    integer_digits.reverse()

    for _ in range(precision):
        fractional *= 60
        digi, fractional = divmod(fractional, 1)
        fractional_digits.append(int(digi))
    return integer_digits, fractional_digits


def numsys_revcast(base, integer_digits, fractional_digits):
    integer = 0
    for ix, digi in enumerate(reversed(integer_digits)):
        integer += base ** ix * digi
    if not fractional_digits:
        return integer
    fractional = 0.
    for ix, digi in enumerate(fractional_digits):
        fractional += (1. / 60) ** (ix + 1) * digi
    return integer + fractional


def overlap_size(v1, v2, u1, u2):
    """
    Calculate overlap size of intervals (v1, v2) and (u1, u2)
    """
    alen = abs(v1 - v2)
    blen = abs(u1 - u2)
    wide = max(v1, v2, u1, u2) - min(v1, v2, u1, u2)
    return alen + blen - wide


def grid_align(v, unitsize):
    unitsize = float(unitsize)
    v = round(v / unitsize) * unitsize
    if round(unitsize) == unitsize:
        return int(v)
    return v


def metric_prefix(number):
    """
    >>> metric_prefix(10 ** 11) 
    (100.0, 'G', 9)
    >>> metric_prefix(.1 ** 5)
    (10.000000000000002, 'u', -6)
    
    :param number: 
    :return: (a, prefix, b)
    number = a * (10 ^ b)
    """
    ix, prefix = 0, ''
    if number >= 1000:
        for ix, prefix in enumerate('kMGTPEZY'):
            number /= 1000.
            if number <= 1000:
                return number, prefix, 3 * (1 + ix)
        return number, prefix, 3 * (1 + ix)
    if number <= .001:
        for ix, prefix in enumerate('munp'):
            number *= 1000.
            if number >= 1:
                return number, prefix, -3 * (1 + ix)
        return number, prefix, -3 * (1 + ix)
    return number, '', 0


def human_filesize(number):
    """
    Human readable file size unit
    :param number: how many bytes
    :return: (num, unit)
    """
    units = ["bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    for unit in units:
        if number < 10000 or unit == "YB":
            return number, unit
        else:
            number /= 1024.0
            # to next loop, no return!



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import collections
import datetime
import functools
import math
import re
import sys
import time
from datetime import timedelta
from typing import NamedTuple

from joker.cast import want_unicode

_sexagesimal_chars = '0123456789aAbBcCdDeEfFgGhHiIjJkKlLmMnNoOpPqQrRsStTuUvVwWxXyY'
_sexagesimal_remap = {ic[1]: ic[0] for ic in enumerate(_sexagesimal_chars)}


def sexagesimal_format(num, precision=0):
    from joker.cast.numeric import numsys_cast

    idigits, fdigits = numsys_cast(num, 60, precision)
    rs = ''.join(_sexagesimal_chars[i] for i in idigits) or '0'

    if isinstance(num, int) and not precision:
        return rs
    return rs + '.' + ''.join(_sexagesimal_chars[i] for i in fdigits)


def sexagesimal_parse(numstr):
    from joker.cast.numeric import numsys_revcast

    if '.' not in numstr:
        digits = [_sexagesimal_remap[c] for c in numstr]
        return numsys_revcast(60, digits, [])

    parts = numstr.split('.', 1)
    idigits = [_sexagesimal_remap[c] for c in parts[0]]
    fdigits = [_sexagesimal_remap[c] for c in parts[1]]
    return numsys_revcast(60, idigits, fdigits or [0])


def start_of_this_year() -> datetime.datetime:
    today = datetime.date.today()
    return datetime.datetime(today.year, 1, 1)


def start_of_this_month() -> datetime.datetime:
    today = datetime.date.today()
    return datetime.datetime(today.year, today.month, 1)


def start_of_today() -> datetime.datetime:
    today = datetime.date.today()
    return datetime.datetime(today.year, today.month, today.day)


def seconds_to_hms(seconds):
    """
    >>> seconds_to_hms(4000)
    (1, 6, 40)
    :return:
    """
    # https://stackoverflow.com/a/775075/2925169
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return h, m, s


def _smart_time_parse(s):
    """
    :param s: (str)
    :return:  [seconds, minutes, hours]
    """
    if ':' in s:
        parts = [float(x or 0) for x in s.split(':')][:3]
        parts.reverse()
        # len(parts) must be 2 or 3
        if len(parts) == 2:
            parts.append(0)
    else:
        sc = s
        parts = []
        for _ in range(3):
            parts.append(int(sc[-2:] or 0))
            sc = sc[:-2]
    return parts


def smart_time_parse(s):
    sec, m, h = _smart_time_parse(s)
    return datetime.time(h, m, sec)


def smart_time_parse_to_seconds(s):
    """
    >>> smart_time_parse_to_seconds('2:3')  # 1min 2sec
    123
    >>> smart_time_parse_to_seconds('1:1:2')  # 1hour 1min 2sec
    3662
    >>> smart_time_parse_to_seconds('1::2')  # 1hour 2sec
    3602
    >>> smart_time_parse_to_seconds('1::')  # 1hour
    3600
    >>> smart_time_parse_to_seconds('10102')  # 1hour 1min 2sec
    3662
    >>> smart_time_parse_to_seconds('200')  # 2min
    120
    :param s: a string representing time
    :return:
    """
    seconds = 0
    for i, x in enumerate(_smart_time_parse(s)):
        seconds += x * 60 ** i
    return seconds


parse_time_to_seconds = smart_time_parse_to_seconds


def smart_time_parse_to_timedelta(s):
    seconds = smart_time_parse_to_seconds(s)
    return datetime.timedelta(seconds=seconds)


def smart_date_parse(s):
    """  
    >>> smart_date_parse(0)
    datetime.date(2017, 5, 5) 
    >>> smart_date_parse('today')
    datetime.date(2017, 5, 5) 
    
    >>> smart_date_parse(-1)
    datetime.date(2017, 5, 4) 
    >>> smart_date_parse('yesterday')
    datetime.date(2017, 5, 4)
    
    >>> smart_date_parse(1)
    datetime.date(2017, 5, 6) 
    >>> smart_date_parse('tomorrow')
    datetime.date(2017, 5, 6) 
    
    >>> smart_date_parse(datetime.date.today())
    datetime.date(2017, 5, 5) 
    >>> smart_date_parse(datetime.datetime.now())
    datetime.date(2017, 5, 5) 
    
    >>> smart_date_parse('20170505')
    datetime.date(2017, 5, 5) 
    >>> smart_date_parse('2017-05-06')
    datetime.date(2017, 5, 6) 
    >>> smart_date_parse('05-06')
    datetime.date(2017, 5, 6) 
    >>> smart_date_parse('0506')
    datetime.date(2017, 5, 6) 
    
    :param s: a string representing a date
    :return: a datetime.date instance
    """
    day = datetime.timedelta(days=1)
    today = datetime.date.today()
    if isinstance(s, int):
        return today + s * day
    elif isinstance(s, datetime.date):
        return s
    elif isinstance(s, datetime.datetime):
        return s.date()
    elif isinstance(s, str):
        s = want_unicode(s).lower()
    else:
        t = s.__class__.__name__
        raise TypeError('cannot convert type {} to date'.format(t))

    if s == 'today':
        return smart_date_parse(0)
    if s == 'yesterday':
        return smart_date_parse(-1)
    if s == 'tomorrow':
        return smart_date_parse(1)

    if re.match(r'\d{8}$', s):
        return datetime.datetime.strptime(s, '%Y%m%d').date()

    if re.match(r'\d{4}$', s):
        s = '{}{}'.format(today.year, s)
        return datetime.datetime.strptime(s, '%Y%m%d').date()

    if re.match(r'\d{4}-\d{1,2}-\d{1,2}$', s):
        return datetime.datetime.strptime(s, '%Y-%m-%d').date()

    if re.match(r'\d{1,2}-\d{1,2}$', s):
        s = '{}-{}'.format(today.year, s)
        return datetime.datetime.strptime(s, '%Y-%m-%d').date()
    raise ValueError('unknow date format')


easy_date = smart_date_parse


def time_format(fmt=None, tm=None):
    if fmt is None:
        fmt = '%y%m%d-%H%M%S'
    if tm is None:
        tm = datetime.datetime.now()
    return tm.strftime(fmt)


def date_range(start, stop=0, step=1):
    """
    >>> list(date_range(-3, 0))  # last 3 days
    [datetime.date(2017, 5, 2),
     datetime.date(2017, 5, 3),
     datetime.date(2017, 5, 4)]
     
    :param start: 
    :param stop: 
    :param step: 
    :return: 
    """
    start = smart_date_parse(start)
    stop = smart_date_parse(stop)
    delta = datetime.timedelta(days=step)
    while (start - stop).total_seconds() * step < 0:
        yield start
        start += delta


class Timer:
    def __init__(self, name: str = '', offset: float = 0.):
        self.name = name
        self.started_at = time.time() + offset
        self.entered_at = None

    def __repr__(self):
        parts = self.__class__.__name__, repr(self.name), -self.seconds
        return '{}({}, {})'.format(*parts)

    def __str__(self):
        tmr = 'Timer {}'.format(self.name).strip()
        if self.entered_at is None:
            return '{}: {}'.format(tmr, self.seconds)
        since_entering = round(time.time() - self.entered_at, 3)
        return '{}: {}, {}'.format(tmr, self.seconds, since_entering)

    def __enter__(self):
        self.entered_at = time.time()
        return self

    def __exit__(self, typ, value, traceback):
        print(self, file=sys.stderr)
        self.entered_at = None

    @property
    def seconds(self):
        diff = time.time() - self.started_at
        return round(diff, 3)

    def as_json_serializable(self):
        return self.seconds


def timed(func):
    """A decorator reporting execution time"""

    @functools.wraps(func)
    def retfunc(*args, **kwargs):
        name = getattr(func, '__name__', None) or func.__class__.__name__
        with Timer(name + '()'):
            return func(*args, **kwargs)

    return retfunc


class TimeMachine(object):
    def __init__(self, start=None, speed=None):
        self._initpoint = datetime.datetime.now()
        self._imaginary = not (start or speed)
        self._start = start or self._initpoint
        self._speed = speed or 1.

    def now(self):
        if not self._imaginary:
            return datetime.datetime.now()
        delta = datetime.datetime.now() - self._initpoint
        return self._start + delta * self._speed

    @staticmethod
    def convert_time_to_timedelta(t):
        return datetime.timedelta(
            hours=t.hour, minutes=t.minute, seconds=t.second,
            microseconds=t.microsecond
        )


class TimeSlicer(TimeMachine):
    EPOCH = datetime.datetime(1970, 1, 1)

    def __init__(self, start=None, speed=None, ts_size=600):
        super(TimeSlicer, self).__init__(start, speed)
        self._ts_delta = datetime.timedelta(seconds=ts_size)

    def get_current_timeslice(self, relative=True):
        dt = self.now()
        return self.convert_datetime_to_timeslice(dt, relative=relative)

    @staticmethod
    def guess_slice_size(ts):
        """
        assuming slice size a multiple of 60
        :param ts:
        :return:
        """
        t = time.time()
        return int(round(t / ts / 60.)) * 60

    def convert_timeslice_to_time(self, ts):
        dt = self.EPOCH + self._ts_delta * ts
        return dt.time()

    def convert_timeslice_to_datetime(self, ts, relative=True):
        """
        If relative is true, given ts is considered relative to
        00:00:00 of today.

        :param ts: an integer, serial number of timeslice
        :param relative: bool
        :return: a datetime instance
        """
        if not relative:
            return self.EPOCH + self._ts_delta * ts
        d = self.now().date()
        t = self.convert_timeslice_to_time(ts)
        return datetime.datetime.combine(d, t)

    def convert_time_to_timeslice(self, t):
        delta = self.convert_time_to_timedelta(t)
        # use .total_seconds to be compat with python 2.x
        return int(delta.total_seconds() / self._ts_delta.total_seconds())

    def convert_datetime_to_timeslice(self, dt, relative=True):
        if relative:
            return self.convert_time_to_timeslice(dt.time())
        delta = dt - self.EPOCH
        # use .total_seconds to be compat with python 2.x
        return int(delta.total_seconds() / self._ts_delta.total_seconds())


class Year(object):
    __slots__ = ['val']

    def __init__(self, num, trad=True):
        # 0 => 1 AD, 1 => 1 AD, -1 => 1 BC
        self.val = num + 1 if trad and num < 1 else num

    def __repr__(self):
        c = self.__class__.__name__
        v = self.val
        num = self.val - 1 if self.val < 1 else self.val
        return '{}({})'.format(c, num)

    def __str__(self):
        if self.val > 0:
            return str(self.val)
        return '{} BC'.format(1 - self.val)

    def __sub__(self, other):
        if isinstance(other, Year):
            return self.val - other.val
        if isinstance(other, int):
            return Year(self.val - other, trad=False)
        return NotImplemented

    def __isub__(self, other):
        if isinstance(other, int):
            self.val -= other
            return self
        raise TypeError('unsupported operation')

    def __add__(self, other):
        if isinstance(other, int):
            return Year(self.val + other, trad=False)
        raise TypeError('unsupported operation')

    def __iadd__(self, other):
        if isinstance(other, int):
            self.val += other
            return self
        raise TypeError('unsupported operation')

    def __eq__(self, other):
        return self.val == other.val

    @classmethod
    def parse(cls, s):
        try:
            return cls(int(s))
        except ValueError:
            pass
        parts = re.split(r'(\d+)', s)
        if parts[0] and parts[-1]:
            raise ValueError("bad format: '{}'".format(s))
        affix = parts[0] or parts[-1]
        affix = affix.strip()
        if affix.upper() in {'BC', 'B.C.', 'BCE', 'B.C.E.'}:
            return cls(-int(parts[1]))
        if affix.upper() in {'AD', 'A.D.', 'CE', 'C.E.'}:
            return cls(int(parts[1]))
        raise ValueError("bad token: '{}'".format(affix))


def _n_minutes_floor(n: int, dt: datetime.datetime):
    minutes = math.floor(dt.minute / n) * n
    return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, minutes)


def _n_minutes_ceil(n: int, dt: datetime.datetime):
    dti = datetime.datetime(dt.year, dt.month, dt.day, dt.hour)
    # minutes can be 60
    minutes = math.ceil(dt.minute / n + dt.second / 60. / n) * n
    return dti + timedelta(minutes=minutes)


def quarter_floor(dt: datetime.datetime):
    return _n_minutes_floor(15, dt)


def quarter_ceil(dt: datetime.datetime):
    return _n_minutes_ceil(15, dt)


def quintminute_floor(dt: datetime.datetime):
    return _n_minutes_floor(5, dt)


def quintminute_ceil(dt: datetime.datetime):
    return _n_minutes_ceil(5, dt)


class TimeSpan(NamedTuple):
    a: datetime.datetime
    b: datetime.datetime

    @classmethod
    def from_timestamps(cls, a: float, b: float):
        return cls(
            datetime.datetime.fromtimestamp(a),
            datetime.datetime.fromtimestamp(b)
        )

    @property
    def duration(self):
        delta = self.b - self.a
        return delta.total_seconds()

    def __contains__(self, v: datetime.datetime) -> bool:
        return self.a <= v <= self.b

    @property
    def as_dict(self):
        return {'a': self.a, 'b': self.b}

    @property
    def as_json_serializable(self):
        return {
            'a': self.a.strftime("%Y-%m-%d %H:%M:%S"),
            'b': self.b.strftime("%Y-%m-%d %H:%M:%S"),
        }

    @property
    def as_mongo_filter(self):
        return {'$gte': self.a, '$lte': self.b}

    @property
    def as_timestamps(self):
        return self.a.timestamp(), self.b.timestamp()

    def fmt_for_human(self):
        if self.a.date() == self.b.date():
            return f'{self.a:%Y-%m-%d %H:%M:%S} ~ {self.b:%H:%M:%S}'
        return f'{self.a:%Y-%m-%d %H:%M:%S} ~ {self.b:%Y-%m-%d %H:%M:%S}'

    def expand(self, a_seconds: int, b_seconds: int = None):
        if b_seconds is None:
            b_seconds = a_seconds
        a_delta = timedelta(seconds=a_seconds)
        b_delta = timedelta(seconds=b_seconds)
        return TimeSpan(self.a - a_delta, self.b + b_delta)

    def expand_to_quarters(self):
        return TimeSpan(quarter_floor(self.a), quarter_ceil(self.b))

    def expand_to_quintminutes(self):
        return TimeSpan(quintminute_floor(self.a), quintminute_ceil(self.b))

    def iter_quarter_time_spans(self):
        whole = self.expand_to_quarters()
        queue = collections.deque([whole.a], maxlen=2)
        delta = timedelta(minutes=15)
        while queue[-1] < whole.b:
            queue.append(queue[-1] + delta)
            yield TimeSpan(*queue)

    def iter_quintminute_time_spans(self):
        whole = self.expand_to_quarters()
        queue = collections.deque([whole.a], maxlen=2)
        delta = timedelta(minutes=5)
        while queue[-1] < whole.b:
            queue.append(queue[-1] + delta)
            yield TimeSpan(*queue)



#!/usr/bin/env python3
# coding: utf-8

import itertools


def flatten(tup):
    """
    >>> flatten([100])
    100
    >>> flatten([100, 200])
    [100, 200]
    :param tup: a tuple or list
    """
    if isinstance(tup, (tuple, list)):
        if len(tup) == 1:
            return tup[0]
        if len(tup) == 0:
            return None
    return tup


def unflatten(obj):
    """
    >>> unflatten(100)
    (100,)
    >>> unflatten(tuple([100]))
    (100,)
    :param obj:
    :return: a tuple or list
    """
    if isinstance(obj, (tuple, list)):
        return obj
    return obj,


def reusable(records):
    """
    >>> reusable(i for i in range(5))
    [0, 1, 2, 3, 4]
    >>> reusable([0, 1, 2, 3, 4])
    [0, 1, 2, 3, 4]
    :param records: an iterable
    :return: a list or tuple
    """
    if isinstance(records, (list, tuple)):
        return records
    return list(records)


def chunkwize(chunksize, iterable):
    """
    >>> list(chunkwize(5, range(14)))
    [[0, 1, 2, 3, 4], 
     [5, 6, 7, 8, 9], 
     [10, 11, 12, 13]]
     
    :param chunksize: integer
    :param iterable:
    """
    chunksize = int(chunksize)
    chunk = []
    for item in iterable:
        if len(chunk) >= chunksize:
            yield chunk
            chunk = []
        chunk.append(item)
    yield chunk


def chunkwize_split(chunksize, sequence):
    n = len(sequence)
    chunksize = chunksize or n
    return [sequence[i: i + chunksize] for i in range(0, n, chunksize)]


def chunkwize_parallel(chunksize, *sequences):
    """
    >>> s1 = '1234567890'
    >>> s2 = 'abcdefghijk'
    >>> list(chunkwize_parallel(4, s1, s2))
    [['1234', 'abcd'], ['5678', 'efgh'], ['90', 'ijk']]
    
    :param chunksize: integer
    :param sequences: tuple of strings or lists (must support slicing!)
    """
    chunksize = int(chunksize)
    for i in itertools.count(0):
        r = [s[i * chunksize:(i + 1) * chunksize] for s in sequences]
        if any(r):
            yield r
        else:
            break


def numseries_segment(wsize, iterable):
    """
    >>> list(numseries_segment(5, [1, 3, 5, 7]))
    [[1, 3], [5, 7]]

    :param wsize: integer
    :param iterable: must be asc ordered
    :return: a list of lists
    """
    wsize = float(wsize)
    count = 0
    chunk = list()
    for num in iterable:
        idx = int(num / wsize)
        while idx > count:
            yield chunk
            chunk = list()
            count += 1
        chunk.append(num)
    yield chunk


def all_combinations(iterable):
    """
    >>> list(all_combinations('abcd'))
    [(),
     ('a',),
     ('b',),
     ('c',),
     ('d',),
     ('a', 'b'),
     ('a', 'c'),
     ('a', 'd'),
     ('b', 'c'),
     ('b', 'd'),
     ('c', 'd'),
     ('a', 'b', 'c'),
     ('a', 'b', 'd'),
     ('a', 'c', 'd'),
     ('b', 'c', 'd'),
     ('a', 'b', 'c', 'd')]
    """
    items = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(items, i) for i in range(1 + len(items)))


def cumsum(nums):
    import operator
    return itertools.accumulate(nums, operator.add)


def window_sum(wsize, numbers):
    """
    >>> nums = [1, 10, 100, 1000, 10000, 100000]
    >>> list(window_sum(3, nums))
    [111, 1110, 11100, 111000]
    
    :param wsize: integer, size of the moving window
    :param numbers: an iterable of numbers
    """
    import collections
    queue = collections.deque(maxlen=wsize)
    for num in numbers:
        queue.append(num)
        if len(queue) >= wsize:
            yield sum(queue)


def alternate(*iterables, **kwargs):
    """
    :param iterables:
    >>> ''.join(list(alternate('ABCD', 'abcde')))
    'AaBbCcDde'
    >>> ''.join(list(alternate('ABCD', 'abcde', fill='_')))
    'AaBbCcDd_e' 
    """
    _void = object()
    fill = kwargs.get('fill', _void)
    alt = itertools.chain(*itertools.zip_longest(*iterables, fillvalue=fill))
    for item in alt:
        if item is not _void:
            yield item


def until_convergent(func, start, n=1000):
    import collections
    que = collections.deque([start], maxlen=2)

    # allow n = float('inf')
    cnt = itertools.count(n, -1)
    while next(cnt) > 0:
        que.append(func(que[0]))
        if que[0] == que[1]:
            return que[0]


def split(iterable, func):
    """
    :param iterable: 
    :param func: func(x) => True if x is a header (footer)
    :return: an iterable if 3-tuples: (header, footer, elements)
    """
    header = None
    elements = []
    for x in iterable:
        if not func(x):
            elements.append(x)
            continue
        if elements or header is not None:
            yield header, x, elements
        header = x
        elements = []
    if elements or header is not None:
        yield header, None, elements


def generic_split(iterable, func):
    header = None
    elements = func()
    for x in iterable:
        container = func(x)
        if container is None:
            elements.append(x)
            continue
        if elements or header is not None:
            yield header, x, elements
        header = x
        elements = container
    if elements or header is not None:
        yield header, None, elements



#!/usr/bin/env python3
# coding: utf-8

__version__ = '0.5.2'


def regular_cast(original, *try_funcs):
    """
    >>> regular_cast('12.3', int, float)
    12.3
    >>> regular_cast('12.3a', int, float)
    '12.3a'
    >>> regular_cast('12.3a', int, float, 0)
    0
    """
    for atmpt in try_funcs:
        if not callable(atmpt):
            return atmpt
        try:
            return atmpt(original)
        except (TypeError, ValueError):
            pass
    return original


def regular_lookup(dictlike, *try_keys):
    for key in try_keys:
        try:
            return dictlike[key]
        except LookupError:
            pass


def regular_attr_lookup(dictlike, *try_attrnames):
    for key in try_attrnames:
        try:
            return getattr(dictlike, key)
        except AttributeError:
            pass


def cache_lookup(dictlike, key, default, *args, **kwargs):
    try:
        return dictlike[key]
    except LookupError:
        pass
    if callable(default):
        value = default(*args, **kwargs)
    else:
        value = default
    dictlike[key] = value
    return value


def smart_cast(value, default):
    """
    Cast to the same type as `default`;
    if fail, return default
    :param value:
    :param default:
    :return:
    """
    func = type(default)
    try:
        return func(value)
    except (TypeError, ValueError):
        return default


def numerify(s):
    return regular_cast(s, int, float)


def want_bytes(s, **kwargs):
    """
    :param s: 
    :param kwargs: key word arguments passed to str.encode(..)
    :return: 
    """
    if not isinstance(s, bytes):
        s = s.encode(**kwargs)
    return s


def want_str(s, **kwargs):
    """
    :param s:
    :param kwargs: key word arguments passed to s.decode(..)
    :return:
    """
    if not isinstance(s, str):
        return s.decode(**kwargs)
    return s


want_unicode = want_str


def represent(obj, params):
    """
    :param obj:
    :param params: a dict or list
    """
    c = obj.__class__.__name__
    if isinstance(params, dict):
        parts = ('{}={}'.format(k, repr(v)) for k, v in params.items())
    else:
        parts = ('{}={}'.format(k, repr(getattr(obj, k))) for k in params)
    s = ', '.join(parts)
    return '<{}({}) at {}>'.format(c, s, hex(id(obj)))



#!/usr/bin/env python3
# coding: utf-8

def noop(*_, **__):
    pass


def default_func(*pargs, **_):
    """
    As a placeholder function, it works like lambda x: x.
    As a method of a class, it returns cls or self.
    :param pargs: positional arguments
    :param _: keyword arguments, ignored
    :return: 1st positional arguments
    """
    return pargs[0]


def adaptive_call(entry):
    """
    >>> import sys
    >>> entr = [print, ['a', 'b'], {'file': sys.stderr}]
    >>> adaptive_call(entr)

    :param entry: an iterable or a callable
    :return:
    """
    pargs = []
    kwargs = dict()
    if callable(entry):
        return entry()

    items = list(entry)
    if not items:
        return items
    if not callable(items[0]):
        raise TypeError('first item of entry must be a callable')

    for x in items[1:]:
        if isinstance(x, (list, tuple)):
            pargs.extend(x)
        elif isinstance(x, dict):
            kwargs.update(x)
        else:
            raise TypeError('params must be a tuple, list or dict')
    return items[0](*pargs, **kwargs)


# TODO: be more general
def printerr(*args, **kwargs):
    import sys
    kwargs.setdefault('file', sys.stderr)
    pargs = []
    for a in args:
        if isinstance(a, BaseException):
            pargs.append(a.__class__.__name__)
            pargs.append(str(a))
            kwargs.setdefault('sep', ': ')
        else:
            pargs.append(a)
    print(*pargs, **kwargs)


def instanciate(cls):
    return cls()


def instanciate_with_foolproof(cls):
    """
    The return class can be called again without error
    """
    if '__call__' not in cls.__dict__:
        cls.__call__ = lambda x: x
    return cls()


class ConstantCallable(object):
    def __init__(self, value):
        self.value = value

    def __call__(self, *args, **kwargs):
        return self.value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self.value)


_always_true = ConstantCallable(True)
_always_false = ConstantCallable(False)


class Void(object):
    """
    Act as 0, False, '', [] 
    """
    __bool__ = _always_false
    __nonzero__ = _always_false
    __add__ = default_func
    __sub__ = default_func
    __radd__ = default_func
    __rsub__ = default_func
    __round__ = default_func
    __truediv__ = default_func
    __floordiv__ = default_func
    __rtruediv__ = default_func
    __rfloordiv__ = default_func
    __rmul__ = default_func
    __gt__ = _always_false
    __ge__ = _always_false
    __lt__ = _always_false
    __le__ = _always_false
    __len__ = ConstantCallable(0)

    def __init__(self, symbol='-'):
        self.symbol = symbol

    def __str__(self):
        return self.symbol

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.symbol))

    def __eq__(self, other):
        if isinstance(other, Void):
            return True
        return False


Object = type('Object', (object,), {})


class Universe(object):
    __contains__ = _always_true
    __iter__ = ConstantCallable(tuple())


class Mu(object):
    __slots__ = ['value']

    # mutable value
    def __init__(self, value):
        self.value = value

    def __call__(self, value):
        self.value = value
        return self


class Glass(object):
    def __getattr__(self, name):
        return globals().get(name)


class KnownKey:
    def __set_name__(self, _, name):
        # __init__ is called before __set_name__
        if not self.name:
            self.name = name

    def __init__(self, name: str = None):
        # __init__ is called before __set_name__
        self.name = name

    def __get__(self, instance, owner):
        if not instance:
            return
        return instance[self.name]


class TransparentWrapper:
    def __init__(self, obj):
        self._obj = obj

    def __getitem__(self, key):
        return self._obj[key]

    def __getattr__(self, name: str):
        return getattr(self._obj, name)


#!/usr/bin/env python3
# coding: utf-8

import collections
import itertools
import math

from joker.cast.syntax import KnownKey, TransparentWrapper

_compat = [KnownKey, TransparentWrapper]


def _is_property(attr):
    if isinstance(attr, property):
        return True
    cn = attr.__class__.__name__
    if cn == 'cached_property':
        return True
    return False


class NovelDict(dict):
    def __getattr__(self, key: str):
        return self[key]

    def __setattr__(self, key: str, value):
        self[key] = value

    def __delattr__(self, key: str):
        del self[key]

    def get_properties(self) -> dict:
        props = {}
        for key, val in self.__class__.__dict__.items():
            if _is_property(val):
                props[key] = getattr(self, key)
        return props

    def get_data_and_properties(self) -> dict:
        data = self.copy()
        data.update(self.get_properties())
        return data


class DefaultOrderedDict(collections.OrderedDict):
    # Source: http://stackoverflow.com/a/6190500/562769
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and
                not isinstance(default_factory, collections.Callable)):
            raise TypeError('first argument must be a callable')
        collections.OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return collections.OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()


class RecursiveDefaultDict(collections.defaultdict):
    def __init__(self, **kwargs):
        super(RecursiveDefaultDict, self).__init__(self.__class__, **kwargs)


class RecursiveCounter(RecursiveDefaultDict):
    def total(self, as_string=False):
        s = 0
        for v in self.values():
            if isinstance(v, int) or isinstance(v, float):
                s += v
            elif isinstance(v, RecursiveCounter):
                s += v.total()
        if as_string:
            s = str(s)
        return s


class _ListWrapper(object):
    def __init__(self, iterable):
        self._items = list(iterable)

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({})'.format(c, self._items)


class Pool(_ListWrapper):
    def __init__(self, iterable):
        super(Pool, self).__init__(iterable)
        self._cycle = itertools.cycle(self._items)

    def shuffle(self):
        import random
        random.shuffle(self._items)
        self._cycle = itertools.cycle(self._items)
        return self

    def take(self, n):
        return list(itertools.islice(self._cycle, n))


class Circular(_ListWrapper):
    """
    >>> c = Circular([0, 1, 2, 3, 4])
    >>> list(c[-1:5])
    [4, 0, 1, 2, 3, 4]
    """

    def __init__(self, iterable):
        super(Circular, self).__init__(iterable)
        assert self._items

    def ix_turn(self, ix):
        if ix is None:
            return
        return len(self._items) - 1 - ix

    def ix_shift(self, *indexes):
        m = min(i for i in indexes if i is not None)
        if m >= 0:
            return indexes
        n = len(self._items)
        shift = math.ceil(-1. * m / n) * n
        shift = int(shift)

        new_indexes = []
        for ix in indexes:
            if ix is None:
                new_indexes.append(None)
            else:
                new_indexes.append(ix + shift)
        return tuple(new_indexes)

    def standardize(self, slc):
        """
        :param slc: a slice instance
        :return: (start, stop, step)  # a tuple
        """
        assert isinstance(slc, slice)
        if slc.step is None or slc.step >= 0:
            return self.ix_shift(slc.start, slc.stop, slc.step)
        return self.ix_shift(
            self.ix_turn(slc.start),
            self.ix_turn(slc.stop),
            -slc.step)

    def __getitem__(self, key):
        n = len(self._items)
        if isinstance(key, int):
            return self._items[key % n]

        if isinstance(key, slice):
            start, stop, step = self.standardize(key)
            if key.step and key.step < 0:
                c_items = itertools.cycle(self._items[::-1])
            else:
                c_items = itertools.cycle(self._items)
            return itertools.islice(c_items, start, stop, step)


class CircularString(object):
    """
    >>> cs = CircularString('0123456789')
    >>> cs[-1:12]
    '9012345678901'
    """

    def __init__(self, string):
        self._string = string
        self._circular = Circular(string)

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({})'.format(c, self._string)

    def __getitem__(self, key):
        return ''.join(list(self._circular[key]))


class CircularReferenceError(ValueError):
    pass


class CircularReferenceDetector(object):
    identset_class = set

    def __init__(self, ident, *ancestors):
        self.ident = ident
        self.ancestors = self.identset_class(ancestors)

    def branch(self, ident):
        if ident in self.ancestors:
            msg = '{} => {}'.format(self.ident, ident)
            raise CircularReferenceError(msg)
        return CircularReferenceDetector(ident, self.ident, *self.ancestors)
